import { Readable } from 'node:stream';
import { Context } from '@eggjs/core';
export interface EggFile {
    field: string;
    filename: string;
    encoding: string;
    mime: string;
    filepath: string;
}
export interface MultipartFileStream extends Readable {
    fields: Record<string, any>;
    filename: string;
    fieldname: string;
    mime: string;
    mimeType: string;
    transferEncoding: string;
    encoding: string;
    truncated: boolean;
}
export interface MultipartOptions {
    autoFields?: boolean;
    /**
     * required file submit, default is true
     */
    requireFile?: boolean;
    /**
     * default charset encoding
     */
    defaultCharset?: string;
    /**
     * compatible with defaultCharset
     * @deprecated use `defaultCharset` instead
     */
    defCharset?: string;
    defaultParamCharset?: string;
    /**
     * compatible with defaultParamCharset
     * @deprecated use `defaultParamCharset` instead
     */
    defParamCharset?: string;
    limits?: {
        fieldNameSize?: number;
        fieldSize?: number;
        fields?: number;
        fileSize?: number;
        files?: number;
        parts?: number;
        headerPairs?: number;
    };
    checkFile?(fieldname: string, file: any, filename: string, encoding: string, mimetype: string): void | Error;
}
export default class MultipartContext extends Context {
    /**
     * create multipart.parts instance, to get separated files.
     * @function Context#multipart
     * @param {Object} [options] - override default multipart configurations
     *  - {Boolean} options.autoFields
     *  - {String} options.defaultCharset
     *  - {String} options.defaultParamCharset
     *  - {Object} options.limits
     *  - {Function} options.checkFile
     * @return {Yieldable | AsyncIterable<Yieldable>} parts
     */
    multipart(options?: MultipartOptions): AsyncIterable<MultipartFileStream>;
    /**
     * save request multipart data and files to `ctx.request`
     * @function Context#saveRequestFiles
     * @param {Object} options - { limits, checkFile, ... }
     */
    saveRequestFiles(options?: MultipartOptions): Promise<void>;
    /**
     * get upload file stream
     * @example
     * ```js
     * const stream = await ctx.getFileStream();
     * // get other fields
     * console.log(stream.fields);
     * ```
     * @function Context#getFileStream
     * @param {Object} options
     *  - {Boolean} options.requireFile - required file submit, default is true
     *  - {String} options.defaultCharset
     *  - {String} options.defaultParamCharset
     *  - {Object} options.limits
     *  - {Function} options.checkFile
     * @return {ReadStream} stream
     * @since 1.0.0
     * @deprecated Not safe enough, use `ctx.multipart()` instead
     */
    getFileStream(options?: MultipartOptions): Promise<MultipartFileStream>;
    /**
     * clean up request tmp files helper
     * @function Context#cleanupRequestFiles
     * @param {Array<String>} [files] - file paths need to cleanup, default is `ctx.request.files`.
     */
    cleanupRequestFiles(files?: EggFile[]): Promise<void>;
}
declare module '@eggjs/core' {
    interface Request {
        /**
         * Files Object Array
         */
        files?: EggFile[];
    }
    interface Context {
        saveRequestFiles(options?: MultipartOptions): Promise<void>;
        getFileStream(options?: MultipartOptions): Promise<MultipartFileStream>;
        cleanupRequestFiles(files?: EggFile[]): Promise<void>;
    }
}
